from typing import List, Literal, Union
from .api_base_model import ApiBaseModelNameLabelDesc
from .schedule_timeline_exit import ScheduleTimelineExit
from .scheduled_instance import ScheduledActivityInstance, ScheduledDecisionInstance
from .timing import Timing


class ScheduleTimeline(ApiBaseModelNameLabelDesc):
    mainTimeline: bool
    entryCondition: str
    entryId: str
    exits: List[ScheduleTimelineExit] = []
    timings: List[Timing] = []
    instances: List[Union[ScheduledActivityInstance, ScheduledDecisionInstance]] = []
    instanceType: Literal["ScheduleTimeline"]



from typing import Union, List, Literal
from .api_base_model import ApiBaseModel
from .study_amendment_reason import StudyAmendmentReason
from .geographic_scope import SubjectEnrollment


class StudyAmendment(ApiBaseModel):
    number: str
    summary: str
    substantialImpact: bool
    primaryReason: StudyAmendmentReason
    secondaryReasons: List[StudyAmendmentReason] = []
    enrollments: List[SubjectEnrollment]
    previousId: Union[str, None] = None
    instanceType: Literal["StudyAmendment"]



import os
import yaml


class FileCache:
    def __init__(self, filepath: str, filename: str):
        self.filename = filename
        self.filepath = filepath

    def exists(self) -> bool:
        return self._file_exists()

    def save(self, data: dict) -> None:
        try:
            if not self._file_exists():
                with open(self._full_filepath(), "w") as f:
                    yaml.dump(data, f, indent=2, sort_keys=True)
        except Exception as e:
            raise Exception(f"Failed to save file '{self._full_filepath()}', {str(e)}")

    def read(self) -> dict:
        try:
            if self._file_exists():
                with open(self._full_filepath()) as f:
                    return yaml.safe_load(f)
            else:
                raise Exception(
                    f"Failed to read file '{self._full_filepath()}', does not exist"
                )
        except Exception as e:
            raise Exception(f"Failed to read file '{self._full_filepath()}', {str(e)}")

    def delete(self) -> None:
        try:
            os.remove(self._full_filepath())
        except Exception:
            pass

    def _file_exists(self) -> bool:
        return os.path.isfile(self._full_filepath())

    def _full_filepath(self) -> str:
        return os.path.join(self.filepath, self.filename)



from typing import Literal
from .api_base_model import ApiBaseModelNameLabelDesc
from .quantity import Quantity
from .administration_duration import AdministrationDuration
from .alias_code import AliasCode


class AgentAdministration(ApiBaseModelNameLabelDesc):
    duration: AdministrationDuration
    dose: Quantity
    route: AliasCode
    frequency: AliasCode
    instanceType: Literal["AgentAdministration"]



import os
import requests
from simple_error_log.errors import Errors
from usdm3.ct.cdisc.library import Library as CtLibrary


class LibraryAPI:
    API_ROOT = "https://api.library.cdisc.org/api/cosmos/v2"

    def __init__(
        self,
        ct_library: CtLibrary,
    ):
        self._errors = Errors()
        self._ct_library = ct_library
        self._api_key = os.getenv("CDISC_API_KEY")
        if not self._api_key:
            self._errors.error("Empty CDISC API key")
        self._headers = {"Content-Type": "application/json", "api-key": self._api_key}
        self._package_metadata = {}
        self._package_items = {}
        self._bc_responses = {}
        self._bcs_raw = {}
        self._map = {}

    def refresh(self):
        self._get_package_metadata()
        self._get_package_items()
        self._get_sdtm_bcs()
        self._get_generic_bcs()
        return self._bcs_raw

    @property
    def errors(self) -> Errors:
        return self._errors

    @property
    def valid(self) -> bool:
        return self._errors.error_count()

    def _get_package_metadata(self):
        urls = {
            "generic": "/mdr/bc/packages",
            "sdtm": "/mdr/specializations/sdtm/packages",
        }
        for url_type, url in urls.items():
            try:
                api_url = self._url(url)
                self._errors.info(f"Processing package metadata for '{api_url}'")
                raw = requests.get(api_url, headers=self._headers)
                response = raw.json()
                print("#", end="", flush=True)
                self._package_metadata[url_type] = response["_links"]["packages"]
            except Exception as e:
                self._errors.exception(
                    f"Failed to retrieve CDISC BC package metadata from '{api_url}'", e
                )

    def _get_package_items(self) -> dict:
        for package_type in ["sdtm", "generic"]:
            self._package_items[package_type] = {}
            for package in self._package_metadata[package_type]:
                self._get_package(package, package_type)

    def _get_package(self, package, package_type):
        try:
            response_field = {
                "sdtm": "datasetSpecializations",
                "generic": "biomedicalConcepts",
            }
            api_url = self._url(package["href"]) if "href" in package else "not set"
            self._errors.info(f"Processing package for '{api_url}'")
            raw = requests.get(api_url, headers=self._headers)
            response = raw.json()
            print("#", end="", flush=True)
            for item in response["_links"][response_field[package_type]]:
                key = item["title"].upper()
                if package_type == "sdtm":
                    self._package_items[package_type][key] = item
                    # map[item['href']] = key
                elif package_type == "generic" and key not in self._package_items:
                    self._package_items[package_type][key] = item
                    self._map[item["href"]] = key
        except Exception as e:
            self._errors.exception(
                f"Failed to retrieve CDISC BC metadata from '{api_url}'", e
            )
            return {}

    def _get_sdtm_bcs(self):
        print(f"\n\nSDTM: {len(self._package_items['sdtm'].keys())}")
        count = 0
        for name, item in self._package_items["sdtm"].items():
            count += 1
            # if count > 5:
            #     break
            print(f"[{count}]", end="", flush=True) if count % 10 == 0 else print(
                ".", end="", flush=True
            )
            self._errors.info(f"Processing SDTM BC '{name}' ...")
            sdtm, generic = self._get_from_url_all(name, item)
            if sdtm:
                bc = self._sdtm_bc_as_usdm(sdtm, generic)
                if bc:
                    if "variables" in sdtm:
                        for item in sdtm["variables"]:
                            property = self._sdtm_bc_property_as_usdm(item, generic)
                            if property:
                                bc["properties"].append(property)
                    self._bcs_raw[name] = bc
                if generic:
                    href = generic["_links"]["self"]["href"]
                    if href in self._map:
                        self._map.pop(href)
                    else:
                        self._errors.info(f"Missing reference when popping {href}")

    def _get_generic_bcs(self) -> dict:
        print(f"\n\nGeneric: {len(self._package_items['generic'].keys())}")
        count = 0
        for name, item in self._package_items["generic"].items():
            count += 1
            # if count > 5:
            #     break
            print(f"[{count}]", end="", flush=True) if count % 10 == 0 else print(
                ".", end="", flush=True
            )
            self._errors.info(f"Processing Generic BC '{name}' ...")
            if self._process_genric_bc(name):
                response = self._get_from_url(item["href"])
                bc = self._generic_bc_as_usdm(response)
                if "dataElementConcepts" in response:
                    for item in response["dataElementConcepts"]:
                        property = self._generic_bc_property_as_usdm(item)
                        if property:
                            bc["properties"].append(property)
                self._bcs_raw[name] = bc

    def _generic_bc_as_usdm(self, api_bc) -> dict:
        concept_code = self._code_object(api_bc["conceptId"], api_bc["shortName"])
        synonyms = api_bc["synonyms"] if "synonyms" in api_bc else []
        return self._biomedical_concept_object(
            api_bc["shortName"],
            api_bc["shortName"],
            synonyms,
            api_bc["_links"]["self"]["href"],
            concept_code,
        )

    def _generic_bc_property_as_usdm(self, property) -> dict:
        concept_code = self._code_object(property["conceptId"], property["shortName"])
        responses = []
        if "exampleSet" in property:
            for example in property["exampleSet"]:
                # term = self._ct_library.preferred_term(example)
                term = None
                if term is not None:
                    code = self._code_object(term["conceptId"], term["preferredTerm"])
                    responses.append(self._response_code_object(code))
        return self._biomedical_concept_property_object(
            property["shortName"],
            property["shortName"],
            property["dataType"],
            responses,
            concept_code,
        )

    def _sdtm_bc_as_usdm(self, sdtm, generic) -> dict:
        try:
            if self._process_sdtm_bc(sdtm["shortName"]):
                role_variable = self._get_role_variable(sdtm)
                if role_variable:
                    if "assignedTerm" in role_variable:
                        if (
                            "conceptId" in role_variable["assignedTerm"]
                            and "value" in role_variable["assignedTerm"]
                        ):
                            concept_code = self._code_object(
                                role_variable["assignedTerm"]["conceptId"],
                                role_variable["assignedTerm"]["value"],
                            )
                        else:
                            self._errors.warning(
                                f"Failed to set BC concept, assigned term path, {sdtm['shortName']}"
                            )
                            concept_code = self._code_object(
                                "No Concept Code",
                                role_variable["assignedTerm"]["value"],
                            )
                    else:
                        self._errors.warning(
                            f"Failed to set BC concept, no assigned term path, {sdtm['shortName']}"
                        )
                        concept_code = self._code_object(
                            generic["conceptId"], generic["shortName"]
                        )
                else:
                    self._errors.warning(
                        f"Failed to set BC concept, no role variable, {sdtm['shortName']}"
                    )
                    concept_code = self._code_object(
                        generic["conceptId"], generic["shortName"]
                    )
                synonyms = generic["synonyms"] if "synonyms" in generic else []
                synonyms.append(generic["shortName"])
                return self._biomedical_concept_object(
                    sdtm["shortName"],
                    sdtm["shortName"],
                    synonyms,
                    sdtm["_links"]["self"]["href"],
                    concept_code,
                )
            else:
                return None
        except Exception as e:
            self._errors.exception(f"Failed to build BC '{sdtm['shortName']}'", e)
            return None

    def _sdtm_bc_property_as_usdm(self, sdtm_property, generic) -> dict:
        try:
            if self._process_property(sdtm_property["name"]):
                if "dataElementConceptId" in sdtm_property:
                    generic_match = self._get_dec_match(
                        generic, sdtm_property["dataElementConceptId"]
                    )
                    if generic_match:
                        concept_code = self._code_object(
                            generic_match["conceptId"], generic_match["shortName"]
                        )
                    else:
                        if (
                            "assignedTerm" in sdtm_property
                            and "conceptId" in sdtm_property["assignedTerm"]
                            and "value" in sdtm_property["assignedTerm"]
                        ):
                            concept_code = self._code_object(
                                sdtm_property["dataElementConceptId"],
                                sdtm_property["name"],
                            )
                        else:
                            self._errors.warning(
                                f"Failed to set property concept, DEC path, {sdtm_property}"
                            )
                            concept_code = self._code_object(
                                sdtm_property["dataElementConceptId"],
                                sdtm_property["name"],
                            )
                else:
                    if "assignedTerm" in sdtm_property:
                        concept_code = self._code_object(
                            sdtm_property["assignedTerm"]["conceptId"],
                            sdtm_property["assignedTerm"]["value"],
                        )
                    else:
                        self._errors.error(
                            f"Failed to set property concept, non DEC path, {sdtm_property}"
                        )
                        concept_code = self._code_object(
                            "No Concept Code", sdtm_property["name"]
                        )
                responses = []
                codes = []
                if "valueList" in sdtm_property:
                    codelist = (
                        sdtm_property["codelist"]["conceptId"]
                        if "codelist" in sdtm_property
                        else None
                    )
                    for value in sdtm_property["valueList"]:
                        term = self._ct_library.preferred_term(value, codelist)
                        if term:
                            code = self._code_object(
                                term["conceptId"], term["preferredTerm"]
                            )
                            codes.append(code)
                        else:
                            term = self._ct_library.submission(value, codelist)
                            if term:
                                code = self._code_object(
                                    term["conceptId"], term["preferredTerm"]
                                )
                                codes.append(code)
                            else:
                                cl = f", code list {sdtm_property['codelist']['conceptId'] if 'codelist' in sdtm_property else '<not defined>'}"
                                self._errors.error(
                                    f"Failed to find submission or preferred term '{value}' {cl}\nProperty:\n{sdtm_property}"
                                )
                for code in codes:
                    response_code = self._response_code_object(code)
                    responses.append(response_code)
                datatype = (
                    sdtm_property["dataType"] if "dataType" in sdtm_property else ""
                )
                return self._biomedical_concept_property_object(
                    sdtm_property["name"],
                    sdtm_property["name"],
                    datatype,
                    responses,
                    concept_code,
                )
            else:
                return None
        except Exception as e:
            self._errors.exception(f"Failed to build property {sdtm_property}", e)
            return None

    def _process_sdtm_bc(self, name):
        if name in [
            "Exclusion Criteria 01",
            "Inclusion Criteria 01",
            "Medical History Prespecified: Alzheimer's Disease",
            "Medical History Prespecified: Confusional Episodes",
            "Medical History Prespecified: Essential Tremor",
            "Medical History Prespecified: Extrapyramidal Features",
            "Medical History Prespecified: Facial Masking",
            "Medical History Prespecified: Rigidity Upper Extremity",
            "Medical History Prespecified: Sensitivity to Neuroleptics",
            "Medical History Prespecified: Visual Hallucinations",
            "TTS Acceptability Survey - Patch Acceptability",
            "TTS Acceptability Survey - Patch Appearance",
            "TTS Acceptability Survey - Patch Durability",
            "TTS Acceptability Survey - Patch Size",
            "Beer Use History",
            "Cigarette History",
            "Cigar History",
            "Coffee Use History",
            "Cola Use History",
            "Distilled Spirits Use History",
            "Pipe History",
            "Tea Use History",
            "Wine Use History",
        ]:
            return False
        return True

    def _process_genric_bc(self, name):
        return True if name.upper() in ["SUBJECT AGE", "RACE", "SEX"] else False

    def _process_property(self, name):
        if name[2:] in [
            "TEST",
            "STRESN",
            "STRESU",
            "STRESC",
            "CLASCD",
            "LOINC",
            "LOT",
            "CAT",
            "SCAT",
            "LLT",
            "LLTCD",
            "HLT",
            "HLTCD",
            "PTCD",
            "BODSYS",
            "BDSYCD",
            "SOC",
            "SOCCD",
            "RLDEV",
        ]:
            return False
        if name in ["EPOCH"]:
            return False
        return True

    def _get_role_variable(self, data):
        try:
            return next(
                (item for item in data["variables"] if item["role"] == "Topic"), None
            )
        except Exception:
            self._errors.warning(f"Failed to find role in {data}")
            return None

    def _get_dec_match(self, data, id):
        try:
            return next(
                (
                    item
                    for item in data["dataElementConcepts"]
                    if item["conceptId"] == id
                ),
                None,
            )
        except Exception:
            self._errors.warning(f"Failed to find DEC in {data} for '{id}'")
            return None

    def _get_from_url_all(self, name, details) -> dict:
        try:
            sdtm_response = self._get_from_url(details["href"])
            generic = sdtm_response["_links"]["parentBiomedicalConcept"]
            generic_response = self._get_from_url(generic["href"])
            return sdtm_response, generic_response
        except Exception as e:
            self._errors.exception(
                f"Failed to retrieve CDISC BC metadata for {name} from '{details['href']}'",
                e,
            )
            return None, None

    def _get_from_url(self, url):
        api_url = self._url(url)
        raw = requests.get(api_url, headers=self._headers)
        result = raw.json()
        return result

    def _url(self, relative_url) -> str:
        return "%s%s" % (self.__class__.API_ROOT, relative_url)

    def _code_object(self, code, decode):
        return {
            "id": "tbd",
            "code": code,
            "codeSystem": self._ct_library.system,
            "codeSystemVersion": self._ct_library.version,
            "decode": decode,
            "instanceType": "Code",
        }

    def _alias_code_object(self, standard_code, aliases):
        return (
            {
                "id": "tbd",
                "standardCode": standard_code,
                "standardCodeAliases": aliases,
                "instanceType": "AliasCode",
            }
            if standard_code
            else None
        )

    def _response_code_object(self, code: dict):
        return {
            "id": "tbd",
            "name": f"RC_{code['code']}",
            "label": "",
            "isEnabled": True,
            "code": code,
            "instanceType": "ResponseCode",
        }

    def _biomedical_concept_property_object(
        self, name, label, datatype, responses, code
    ) -> dict:
        alias_code = self._alias_code_object(code, [])
        return {
            "id": "tbd",
            "name": name,
            "label": label,
            "isRequired": True,
            "isEnabled": True,
            "datatype": datatype,
            "responseCodes": responses,
            "code": alias_code,
            "instanceType": "BiomedicalConceptProperty",
        }

    def _biomedical_concept_object(
        self, name, label, synonyms, reference, code
    ) -> dict:
        alias_code = self._alias_code_object(code, [])
        return {
            "id": "tbd",
            "name": name,
            "label": label,
            "synonyms": synonyms,
            "reference": reference,
            "properties": [],
            "code": alias_code,
            "instanceType": "BiomedicalConcept",
        }



from .rule_template import RuleTemplate
from usdm3.data_store.data_store import DataStore


class RuleDDF00011(RuleTemplate):
    def __init__(self):
        super().__init__(
            "DDF00011",
            RuleTemplate.ERROR,
            'Anchor timings (e.g. type is "Fixed Reference") must be related to a scheduled activity instance via a relativeFromScheduledInstance relationship.',
        )

    def validate(self, config: dict) -> bool:
        data: DataStore = config["data"]
        items = data.instances_by_klass("Timing")
        for item in items:
            if item["type"]["decode"] == "Fixed Reference":
                if "relativeFromScheduledInstanceId" not in item:
                    self._add_failure(
                        "Missing relativeFromScheduledInstance",
                        "Timing",
                        "relativeFromScheduledInstanceId",
                        data.path_by_id(item["id"]),
                    )
        return self._result()



from simple_error_log.error_location import ErrorLocation
from simple_error_log.errors import Errors
from usdm3.ct.cdisc.library import Library as CTLibrary


class ValidationLocation(ErrorLocation):
    def __init__(
        self, rule: str, rule_text: str, klass: str, attribute: str, path: str
    ):
        self.rule = rule
        self.rule_text = rule_text
        self.klass = klass
        self.attribute = attribute
        self.path = path

    def to_dict(self):
        return {
            "rule": self.rule,
            "rule_text": self.rule_text,
            "klass": self.klass,
            "attribute": self.attribute,
            "path": self.path,
        }

    @classmethod
    def headers(self):
        return ["rule", "rule_text", "klass", "attribute", "path"]

    def __str__(self):
        return f"{self.rule} [{self.rule_text}]: {self.klass}.{self.attribute} at {self.path}"


class RuleTemplate:
    """
    Base class for rule templates
    """

    ERROR = Errors.ERROR
    WARNING = Errors.WARNING

    class CTException(Exception):
        pass

    def __init__(self, rule: str, level: int, rule_text: str):
        self._errors = Errors()
        self._rule = rule
        self._level = level
        self._rule_text = rule_text

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")

    def errors(self) -> Errors:
        return self._errors

    def _add_failure(self, message: str, klass: str, attribute: str, path: str):
        location = ValidationLocation(
            self._rule, self._rule_text, klass, attribute, path
        )
        self._errors.add(message, location, self._rule, self._level)

    def _result(self) -> bool:
        return self._errors.count() == 0

    def _ct_check(self, config: dict, klass: str, attribute: str) -> bool:
        data = config["data"]
        ct = config["ct"]
        instances = data.instances_by_klass(klass)
        codes, decodes = self._check_codelist(ct, klass, attribute)
        for instance in instances:
            if attribute in instance:
                items = (
                    instance[attribute]
                    if isinstance(instance[attribute], list)
                    else [instance[attribute]]
                )
                for item in items:
                    # print(f"ITEM: {item}")
                    code = item["code"] if "code" in item else None
                    decode = item["decode"] if "decode" in item else None
                    code_index = self._find_index(codes, code)
                    decode_index = self._find_index(decodes, decode)
                    if code_index is None and decode_index is not None:
                        self._add_failure(
                            f"Invalid code '{code}', the code is not in the codelist",
                            klass,
                            attribute,
                            data.path_by_id(instance["id"]),
                        )
                    elif code_index is not None and decode_index is None:
                        self._add_failure(
                            f"Invalid decode '{decode}', the decode is not in the codelist",
                            klass,
                            attribute,
                            data.path_by_id(instance["id"]),
                        )
                    elif code_index is None and decode_index is None:
                        self._add_failure(
                            f"Invalid code and decode '{code}' and '{decode}', neither the code and decode are in the codelist",
                            klass,
                            attribute,
                            data.path_by_id(instance["id"]),
                        )
                    elif code_index != decode_index:
                        self._add_failure(
                            f"Invalid code and decode pair '{code}' and '{decode}', the code and decode do not match",
                            klass,
                            attribute,
                            data.path_by_id(instance["id"]),
                        )
            else:
                self._add_failure(
                    "Missing attribute",
                    klass,
                    attribute,
                    data.path_by_id(instance["id"]),
                )
        return self._result()

    def _check_codelist(
        self, ct: CTLibrary, klass: str, attribute: str
    ) -> tuple[list[str], list[str]]:
        codelist = ct.klass_and_attribute(klass, attribute)
        if codelist is None:
            raise self.CTException(
                f"Failed to find code list for '{klass}' and '{attribute}'"
            )
        codes, decodes = self._codes_and_decodes(codelist)
        if codes is None:
            raise self.CTException(
                f"Failed to find terms for '{klass}' and '{attribute}'"
            )
        return codes, decodes

    def _codes_and_decodes(self, codelist: dict) -> tuple[list[str], list[str]]:
        if "terms" not in codelist or codelist["terms"] == []:
            return None, None
        codes = [x["conceptId"] for x in codelist["terms"]]
        decodes = [x["preferredTerm"] for x in codelist["terms"]]
        return codes, decodes

    def _find_index(self, items: list[str], value: str) -> int | None:
        try:
            return items.index(value)
        except ValueError:
            return None



import json
from src.usdm3 import USDM3
from tests.rules.helpers import clear_rules_library
from usdm3.__info__ import __package_version__


def test_validate(tmp_path):
    # Create temporary test file
    clear_rules_library()
    test_file = tmp_path / "validate.json"
    with open(test_file, "w") as f:
        json.dump(_expected(), f)
    result = USDM3().validate(test_file)
    assert result.passed_or_not_implemented()


def test_validate_error(tmp_path):
    # Create temporary test file
    clear_rules_library()
    test_file = tmp_path / "validate.json"
    with open(test_file, "w") as f:
        json.dump(_bad(), f)
    result = USDM3().validate(test_file)
    # print(f"RESULT: {result._items['DDF00082']}")
    assert not result.passed_or_not_implemented()


def test_minimum():
    result = USDM3().minimum("Name", "ACME", "1")
    result.study.id = "FAKE-UUID"
    assert result.model_dump() == _expected()


def _bad():
    data = _expected()
    data["study"]["documentedBy"]["id"] = None
    return data


def _expected():
    return {
        "id": "Wrapper_1",
        "study": {
            "description": "",
            "documentedBy": {
                "description": "The study protocol document",
                "id": "StudyProtocolDocument_1",
                "instanceType": "StudyProtocolDocument",
                "label": "Study Protocol",
                "name": "PROTOCOL",
                "versions": [
                    {
                        "childIds": [],
                        "contents": [],
                        "dateValues": [],
                        "id": "StudyProtocolDocumentVersion_1",
                        "instanceType": "StudyProtocolDocumentVersion",
                        "protocolStatus": {
                            "code": "C25425",
                            "codeSystem": "cdisc.org",
                            "codeSystemVersion": "2024-09-27",
                            "decode": "Approved",
                            "id": "Code_3",
                            "instanceType": "Code",
                        },
                        "protocolVersion": "1",
                    },
                ],
            },
            "id": "FAKE-UUID",
            "instanceType": "Study",
            "label": "",
            "name": "Study",
            "versions": [
                {
                    "amendments": [],
                    "businessTherapeuticAreas": [],
                    "dateValues": [],
                    "documentVersionId": "StudyProtocolDocumentVersion_1",
                    "id": "StudyVersion_1",
                    "instanceType": "StudyVersion",
                    "rationale": "To be provided",
                    "studyDesigns": [],
                    "studyIdentifiers": [
                        {
                            "id": "StudyIdentifier_1",
                            "instanceType": "StudyIdentifier",
                            "studyIdentifier": "ACME",
                            "studyIdentifierScope": {
                                "id": "Organization_1",
                                "identifier": "To be provided",
                                "identifierScheme": "To be provided",
                                "instanceType": "Organization",
                                "label": None,
                                "legalAddress": None,
                                "name": "Sponsor",
                                "organizationType": {
                                    "code": "C70793",
                                    "codeSystem": "cdisc.org",
                                    "codeSystemVersion": "2024-09-27",
                                    "decode": "Clinical Study Sponsor",
                                    "id": "Code_2",
                                    "instanceType": "Code",
                                },
                            },
                        },
                    ],
                    "studyPhase": None,
                    "studyType": None,
                    "titles": [
                        {
                            "id": "StudyTitle_1",
                            "instanceType": "StudyTitle",
                            "text": "Name",
                            "type": {
                                "code": "C207616",
                                "codeSystem": "cdisc.org",
                                "codeSystemVersion": "2024-09-27",
                                "decode": "Official Study Title",
                                "id": "Code_1",
                                "instanceType": "Code",
                            },
                        },
                    ],
                    "versionIdentifier": "1",
                },
            ],
        },
        "systemName": "Python USDM3 Package",
        "systemVersion": __package_version__,
        "usdmVersion": "3.0.0",
    }



from typing import List, Literal
from .api_base_model import ApiBaseModelNameLabelDesc
from .code import Code


class StudyArm(ApiBaseModelNameLabelDesc):
    type: Code
    dataOriginDescription: str
    dataOriginType: Code
    populationIds: List[str] = []
    instanceType: Literal["StudyArm"]



from .rule_template import RuleTemplate


class RuleDDF00150(RuleTemplate):
    """
    DDF00150: An encounter type must be specified according to the extensible encounter type (C188728) DDF codelist (e.g. an entry with a code or decode used from the codelist should be consistent with the full entry in the codelist).

    Applies to: Encounter
    Attributes: type
    """

    def __init__(self):
        super().__init__(
            "DDF00150",
            RuleTemplate.ERROR,
            "An encounter type must be specified according to the extensible encounter type (C188728) DDF codelist (e.g. an entry with a code or decode used from the codelist should be consistent with the full entry in the codelist).",
        )

    def validate(self, config: dict) -> bool:
        return self._ct_check(config, "Encounter", "type")



import re


class IdManager:
    def __init__(self, classes: list[str]):
        self._classes = classes
        self._id_index = {}
        self.clear()

    def clear(self):
        for klass in self._classes:
            name = klass if isinstance(klass, str) else klass.__name__
            self._id_index[name] = 0

    def build_id(self, klass):
        klass_name = klass if isinstance(klass, str) else str(klass.__name__)
        self._id_index[klass_name] += 1
        return f"{klass_name}_{self._id_index[klass_name]}"

    def add_id(self, klass_name: str, id: str) -> None:
        if instance := self._find_id_instance(id):
            if instance > self._id_index[klass_name]:
                self._id_index[klass_name] = instance

    def _find_id_instance(self, id_text: str) -> int | None:
        pattern = r"^.+_(\d+)$"
        match = re.match(pattern, id_text)
        return int(match.group(1)) if match else None



from .rule_template import RuleTemplate


class RuleDDF00141(RuleTemplate):
    """
    DDF00141: A planned sex must be specified using the Sex of Participants (C66732) SDTM codelist.

    Applies to: StudyDesignPopulation, StudyCohort
    Attributes: plannedSex
    """

    def __init__(self):
        super().__init__(
            "DDF00141",
            RuleTemplate.ERROR,
            "A planned sex must be specified using the Sex of Participants (C66732) SDTM codelist.",
        )

    def validate(self, config: dict) -> bool:
        pop_result = self._ct_check(config, "StudyDesignPopulation", "plannedSex")
        cohort_result = self._ct_check(config, "StudyCohort", "plannedSex")
        return pop_result or cohort_result



import os
import yaml


class Missing:
    def __init__(self, file_path: str):
        f = open(os.path.join(file_path, "missing_ct.yaml"))
        self._missing_ct = yaml.load(f, Loader=yaml.FullLoader)

    def code_lists(self):
        for response in self._missing_ct:
            yield response



from .rule_template import RuleTemplate


class RuleDDF00116(RuleTemplate):
    """
    DDF00116: A study version's study type must be specified using the Study Type Response (C99077) SDTM codelist.

    Applies to: StudyVersion
    Attributes: studyType
    """

    def __init__(self):
        super().__init__(
            "DDF00116",
            RuleTemplate.ERROR,
            "A study version's study type must be specified using the Study Type Response (C99077) SDTM codelist.",
        )



from typing import Literal, Union
from .api_base_model import ApiBaseModelNameLabelDesc
from .code import Code


class Procedure(ApiBaseModelNameLabelDesc):
    procedureType: str
    code: Code
    studyInterventionId: Union[str, None] = None
    instanceType: Literal["Procedure"]



from typing import Union, List, Literal
from .api_base_model import ApiBaseModelNameLabelDesc
from .code import Code
from .range import Range
from .eligibility_criterion import EligibilityCriterion
from .characteristic import Characteristic


class PopulationDefinition(ApiBaseModelNameLabelDesc):
    includesHealthySubjects: bool
    plannedEnrollmentNumber: Union[Range, None] = None
    plannedCompletionNumber: Union[Range, None] = None
    plannedSex: List[Code] = []
    criteria: List[EligibilityCriterion]
    plannedAge: Union[Range, None] = None
    instanceType: Literal["PopulationDefinition"]


class StudyCohort(PopulationDefinition):
    characteristics: List[Characteristic] = []
    instanceType: Literal["StudyCohort"]


class StudyDesignPopulation(PopulationDefinition):
    cohorts: List[StudyCohort] = []
    instanceType: Literal["StudyDesignPopulation"]



from usdm3.base.id_manager import IdManager


class APIInstance:
    class APIInstanceError(Exception):
        """Custom exception for create errors"""

        pass

    def __init__(self, id_manager: IdManager):
        self._id_manager = id_manager

    def create(self, klass, params):
        try:
            klass_name = klass if isinstance(klass, str) else klass.__name__
            params["id"] = (
                self._id_manager.build_id(klass_name)
                if "id" not in params
                else params["id"]
            )
            params["instanceType"] = klass_name
            return klass(**params)
        except Exception as e:
            raise self.APIInstanceError(
                f"Failed to create {klass_name} instance, details: {e}"
            )



from .rule_template import RuleTemplate


class RuleDDF00121(RuleTemplate):
    """
    DDF00121: A study design's blinding schema must be specified according to the extensible Trial Blinding Schema Response (C66735) SDTM codelist (e.g. an entry with a code or decode used from the codelist should be consistent with the full entry in the codelist).

    Applies to: StudyDesign
    Attributes: blindingSchema
    """

    def __init__(self):
        super().__init__(
            "DDF00121",
            RuleTemplate.ERROR,
            "A study design's blinding schema must be specified according to the extensible Trial Blinding Schema Response (C66735) SDTM codelist (e.g. an entry with a code or decode used from the codelist should be consistent with the full entry in the codelist).",
        )



from .rule_template import RuleTemplate


class RuleDDF00086(RuleTemplate):
    """
    DDF00086: Syntax template text is expected to be HTML formatted.

    Applies to: EligibilityCriterion, Characteristic, Condition, Objective, Endpoint
    Attributes: text
    """

    def __init__(self):
        super().__init__(
            "DDF00086",
            RuleTemplate.WARNING,
            "Syntax template text is expected to be HTML formatted.",
        )



import os
import yaml


class Config:
    def __init__(self, filepath: str, filename: str = "ct_config.yaml"):
        self.filename = filename
        self.filepath = filepath
        f = open(os.path.join(self._full_filepath()))
        self._cdisc_ct_config = yaml.load(f, Loader=yaml.FullLoader)
        self._by_klass_attribute = {}
        self._process()

    def required_code_lists(self) -> list:
        return self._cdisc_ct_config["code_lists"]

    def required_packages(self) -> list:
        return self._cdisc_ct_config["packages"]

    def klass_and_attribute(self, klass, attribute) -> str:
        try:
            klass = klass if isinstance(klass, str) else klass.__name__
            return self._by_klass_attribute[klass][attribute]
        except Exception:
            raise ValueError(
                f"failed to find codelist for class '{klass}' attribute '{attribute}'"
            )

    def _process(self):
        for klass, info in self._cdisc_ct_config["klass_attribute_mapping"].items():
            if klass not in self._by_klass_attribute:
                self._by_klass_attribute[klass] = {}
            for attribute, cl in info.items():
                if attribute not in self._by_klass_attribute[klass]:
                    self._by_klass_attribute[klass][attribute] = cl

    def _full_filepath(self) -> str:
        return os.path.join(self.filepath, self.filename)



from .rule_template import RuleTemplate
from usdm3.__info__ import __model_version__ as model_version


class RuleDDFSDW001(RuleTemplate):
    """
    DDFSDW001: The version in the wrapper should be set to 3.0.0
    Applies to: Wrapper
    Attributes: usdmVersion
    """

    def __init__(self):
        super().__init__(
            "DDFSDW001",
            RuleTemplate.ERROR,
            "The version in the wrapper should be set to 3.0.0",
        )

    def validate(self, config: dict) -> bool:
        return self._validate_version(config, model_version)

    def _validate_version(self, config: dict, version: str) -> bool:
        data = config["data"]
        items = data.instances_by_klass("Wrapper")
        if (
            len(items) == 1
            and "usdmVersion" in items[0]
            and items[0]["usdmVersion"] == version
        ):
            pass
        else:
            self._add_failure(
                f"Invalid version detected, not set to {version}",
                "Wrapper",
                "usdmVersion",
                data.path_by_id("$root"),
            )
        return self._result()



from typing import Literal, Union
from .api_base_model import ApiBaseModel
from .code import Code
from .alias_code import AliasCode
from .quantity import Quantity


class GeographicScope(ApiBaseModel):
    type: Code
    code: Union[AliasCode, None] = None
    instanceType: Literal["GeographicScope"]


class SubjectEnrollment(GeographicScope):
    quantity: Quantity
    instanceType: Literal["SubjectEnrollment"]



from .rule_template import RuleTemplate


class RuleDDF00085(RuleTemplate):
    """
    DDF00085: Narrative content text is expected to be HTML formatted.

    Applies to: NarrativeContent
    Attributes: text
    """

    def __init__(self):
        super().__init__(
            "DDF00085",
            RuleTemplate.WARNING,
            "Narrative content text is expected to be HTML formatted.",
        )



import os
from usdm3.ct.cdisc.library_api import LibraryAPI
from usdm3.ct.cdisc.config.config import Config
from usdm3.ct.cdisc.missing.missing import Missing
from usdm3.ct.cdisc.library_cache.library_cache import LibraryCache


class Library:
    """
    A class to manage CDISC controlled terminology (CT) data.

    This class handles loading, caching, and accessing CDISC controlled terminology,
    including code lists and their associated terms. It can load data from a local
    cache file or fetch it from the CDISC API when needed.
    """

    BASE_PATH = "ct/cdisc"
    USDM = "usdm"
    ALL = "all"

    def __init__(self, root_path: str, type: str = USDM):
        self.system = "http://www.cdisc.org"
        self.version = "2023-12-15"  # Default version.
        self.root_path = root_path
        self._type = (
            type.lower() if type.lower() in [self.USDM, self.ALL] else self.USDM
        )

        self._config = Config(
            os.path.join(self.root_path, self.BASE_PATH, "config")
        )  # Configuration for required code lists and mappings
        self._missing = Missing(
            os.path.join(self.root_path, self.BASE_PATH, "missing")
        )  # Handler for missing/additional code lists
        self._api = LibraryAPI(
            self._config.required_packages()
        )  # Interface to CDISC Library API
        self._cache = LibraryCache(
            os.path.join(self.root_path, self.BASE_PATH, "library_cache"),
            f"library_cache_{self._type}.yaml",
        )  # Cache file handler

        # Data structures to store and index controlled terminology
        self._by_code_list = {}  # Maps concept IDs to complete code list data
        self._by_term = {}  # Maps term concept IDs to parent code list IDs
        self._by_submission = {}  # Maps submission values to parent code list IDs
        self._by_pt = {}  # Maps preferred terms to parent code list IDs

    def load(self) -> None:
        if self._cache.exists():
            self._load_ct()  # Load from cache file
        else:
            self._api.refresh()  # Ensure API connection is fresh
            self._get_usdm_ct() if self._usdm() else self._get_all_ct()  # Fetch from API
            self._cache.save(self._by_code_list)  # Cache the results
        self._add_missing_ct()  # Add any additional required terminology

    def klass_and_attribute(self, klass, attribute) -> dict:
        try:
            concept_id = self._config.klass_and_attribute(klass, attribute)
            return self._by_code_list[concept_id]
        except Exception:
            return None

    def klass_and_attribute_value(
        self, klass: str, attribute: str, value: str
    ) -> tuple[dict, str]:
        try:
            concept_id = self._config.klass_and_attribute(klass, attribute)
            code_list = self._by_code_list[concept_id]
            return self._get_item(code_list, value), code_list["source"][
                "effective_date"
            ]
        except Exception:
            return None, None

    def unit(self, value: str) -> dict:
        try:
            code_list = self._by_code_list["C71620"]
            return self._get_item(code_list, value)
        except Exception:
            return None

    def unit_code_list(self) -> dict:
        return self._by_code_list["C71620"]

    def cl_by_term(self, term_code: str) -> dict:
        try:
            concept_ids = self._by_term[term_code]
            return self._by_code_list[concept_ids[0]]
        except Exception:
            return None

    def submission(self, value, cl=None):
        if value in list(self._by_submission.keys()):
            return self._find_in_collection(
                self._by_submission[value], "submissionValue", value, cl
            )
        else:
            return None

    def preferred_term(self, value, cl=None):
        if value in list(self._by_pt.keys()):
            return self._find_in_collection(
                self._by_pt[value], "preferredTerm", value, cl
            )
        else:
            return None

    def _usdm(self) -> bool:
        return self._type == self.USDM

    def _find_in_collection(self, concepts: list, key: str, value: str, cl: str = None):
        if len(concepts) == 0:
            return None
        elif len(concepts) == 1:
            code_list = self._by_code_list[concepts[0]]
            return next(
                (
                    item
                    for item in code_list["terms"]
                    if item[key].upper() == value.upper()
                ),
                None,
            )
        else:
            if cl and cl in concepts:
                code_list = self._by_code_list[cl]
                return next(
                    (
                        item
                        for item in code_list["terms"]
                        if item[key].upper() == value.upper()
                    ),
                    None,
                )
            else:
                return None

    def _get_item(self, code_list, value) -> dict:
        try:
            for field in ["conceptId", "preferredTerm", "submissionValue"]:
                result = next(
                    (
                        item
                        for item in code_list["terms"]
                        if item[field].upper() == value.upper()
                    ),
                    None,
                )
                if result:
                    return result
            return None
        except Exception:
            return None

    def _get_usdm_ct(self) -> None:
        for item in self._config.required_code_lists():
            print(f"[{item}] ", end="", flush=True)
            response = self._api.code_list(item)
            self._by_code_list[response["conceptId"]] = response
            for item in response["terms"]:
                # Index each term by its various identifiers
                self._check_in_and_add(
                    self._by_term, item["conceptId"], response["conceptId"]
                )
                self._check_in_and_add(
                    self._by_submission, item["submissionValue"], response["conceptId"]
                )
                self._check_in_and_add(
                    self._by_pt, item["preferredTerm"], response["conceptId"]
                )

    def _get_all_ct(self) -> None:
        for package in self._api.all_code_lists():
            length = len(package["code_lists"])
            print(f"\n\nPackage: {package}: {length}\n")
            for index, code_list in enumerate(package["code_lists"]):
                response = self._api.package_code_list(
                    package["package"], package["effective_date"], code_list
                )
                print(f"[{index}]", end="", flush=True) if index % 10 == 0 else print(
                    "#", end="", flush=True
                )
                self._by_code_list[response["conceptId"]] = response
                for item in response["terms"]:
                    # Index each term by its various identifiers
                    self._check_in_and_add(
                        self._by_term, item["conceptId"], response["conceptId"]
                    )
                    self._check_in_and_add(
                        self._by_submission,
                        item["submissionValue"],
                        response["conceptId"],
                    )
                    self._check_in_and_add(
                        self._by_pt, item["preferredTerm"], response["conceptId"]
                    )

    def _load_ct(self) -> None:
        self._by_code_list = self._cache.read()
        for c_code, entry in self._by_code_list.items():
            for item in entry["terms"]:
                # Rebuild indexes from cached data
                self._check_in_and_add(self._by_term, item["conceptId"], c_code)
                self._check_in_and_add(
                    self._by_submission, item["submissionValue"], c_code
                )
                self._check_in_and_add(self._by_pt, item["preferredTerm"], c_code)

    def _add_missing_ct(self) -> None:
        for response in self._missing.code_lists():
            self._by_code_list[response["conceptId"]] = response
            for item in response["terms"]:
                # Index the additional terms
                self._check_in_and_add(
                    self._by_term, item["conceptId"], response["conceptId"]
                )
                self._check_in_and_add(
                    self._by_submission, item["submissionValue"], response["conceptId"]
                )
                self._check_in_and_add(
                    self._by_pt, item["preferredTerm"], response["conceptId"]
                )

    def _check_in_and_add(self, collection: dict, id: str, item: str) -> None:
        if id not in collection:
            collection[id] = []
        collection[id].append(item)



from .rule_template import RuleTemplate


class RuleDDF00148(RuleTemplate):
    """
    DDF00148: An endpoint level must be specified using the endpoint level (C188726) DDF codelist.

    Applies to: Endpoint
    Attributes: level
    """

    def __init__(self):
        super().__init__(
            "DDF00148",
            RuleTemplate.ERROR,
            "An endpoint level must be specified using the endpoint level (C188726) DDF codelist.",
        )

    def validate(self, config: dict) -> bool:
        return self._ct_check(config, "Endpoint", "level")



from .rule_template import RuleTemplate


class RuleDDF00108(RuleTemplate):
    """
    DDF00108: There must be at least one exit defined for each timeline (i.e., at least one instance of StudyTimelineExit linked via the 'exits' relationship).

    Applies to: ScheduleTimeline
    Attributes: exits
    """

    def __init__(self):
        super().__init__(
            "DDF00108",
            RuleTemplate.ERROR,
            "There must be at least one exit defined for each timeline (i.e., at least one instance of StudyTimelineExit linked via the 'exits' relationship).",
        )

    def validate(self, config: dict) -> bool:
        data = config["data"]
        items = data.instances_by_klass("StudyTimeline")
        for item in items:
            if "exits" in item:
                if len(item["exits"]) == 0:
                    self._add_failure(
                        "No exits defined for timeline",
                        "StudyTimeline",
                        "exits",
                        data.path_by_id(item["id"]),
                    )
            else:
                self._add_failure(
                    "Missing exits",
                    "StudyTimeline",
                    "exits",
                    data.path_by_id(item["id"]),
                )
        return self._errors.count() == 0



from typing import Literal
from .api_base_model import ApiBaseModelNameLabelDesc


class IntercurrentEvent(ApiBaseModelNameLabelDesc):
    strategy: str
    instanceType: Literal["IntercurrentEvent"]



from .rule_template import RuleTemplate


class RuleDDF00147(RuleTemplate):
    """
    DDF00147: An objective level must be specified using the objective level (C188725) DDF codelist.

    Applies to: Objective
    Attributes: level
    """

    def __init__(self):
        super().__init__(
            "DDF00147",
            RuleTemplate.ERROR,
            "An objective level must be specified using the objective level (C188725) DDF codelist.",
        )

    def validate(self, config: dict) -> bool:
        return self._ct_check(config, "Objective", "level")



from typing import List, Literal
from .api_base_model import ApiBaseModelNameLabelDesc
from .code import Code


class Indication(ApiBaseModelNameLabelDesc):
    codes: List[Code] = []
    isRareDisease: bool
    instanceType: Literal["Indication"]



from .rule_template import RuleTemplate


class RuleDDF00142(RuleTemplate):
    """
    DDF00142: A governance date type must be specified according to the extensible governance date type (C207413) DDF codelist (e.g. an entry with a code or decode used from the codelist should be consistent with the full entry in the codelist).

    Applies to: GovernanceDate
    Attributes: type
    """

    def __init__(self):
        super().__init__(
            "DDF00142",
            RuleTemplate.ERROR,
            "A governance date type must be specified according to the extensible governance date type (C207413) DDF codelist (e.g. an entry with a code or decode used from the codelist should be consistent with the full entry in the codelist).",
        )

    def validate(self, config: dict) -> bool:
        return self._ct_check(config, "GovernanceDate", "type")



import os
import re
import requests


class LibraryAPI:
    API_ROOT = "https://api.library.cdisc.org/api"

    def __init__(self, packages: list) -> None:
        self._packages = None
        self._package_list = packages
        self._headers = {
            "Content-Type": "application/json",
            "api-key": os.environ.get("CDISC_API_KEY"),
        }

    def refresh(self) -> None:
        self._packages = self._get_packages()

    def code_list(self, c_code: str) -> dict | None:
        for package in self._package_list:
            version = self._package_version(package)
            if version:
                package_full_name = f"{package}-{version}"
                api_url = self._url(
                    f"/mdr/ct/packages/{package_full_name}/codelists/{c_code}"
                )
                raw = requests.get(api_url, headers=self._headers)
                if raw.status_code == 200:
                    response = raw.json()
                    response.pop("_links", None)
                    response["source"] = {"effective_date": version, "package": package}
                    return response
        return None

    def all_code_lists(self) -> list:
        results = []
        for package in self._package_list:
            version = self._package_version(package)
            if version:
                package_full_name = f"{package}-{version}"
                api_url = self._url(f"/mdr/ct/packages/{package_full_name}/codelists")
                raw = requests.get(api_url, headers=self._headers)
                if raw.status_code == 200:
                    response = raw.json()
                    result = {
                        "effective_date": version,
                        "package": package,
                        "code_lists": [],
                    }
                    for item in response["_links"]["codelists"]:
                        href = item["href"]
                        result["code_lists"].append(href.split("/")[-1])
                    results.append(result)
        return results

    def package_code_list(self, package: str, version: str, c_code: str) -> dict:
        package_full_name = f"{package}-{version}"
        api_url = self._url(f"/mdr/ct/packages/{package_full_name}/codelists/{c_code}")
        raw = requests.get(api_url, headers=self._headers)
        if raw.status_code == 200:
            response = raw.json()
            response.pop("_links", None)
            response["source"] = {"effective_date": version, "package": package}
            return response
        else:
            return None

    def _get_packages(self) -> dict | None:
        packages = {}
        api_url = self._url("/mdr/ct/packages")
        raw = requests.get(api_url, headers=self._headers)
        if raw.status_code == 200:
            response = raw.json()
            for item in response["_links"]["packages"]:
                name = self._extract_ct_name(item["href"])
                effective_date = self._extract_effective_date(item["title"])
                if name and effective_date:
                    if name not in packages:
                        packages[name] = []
                    packages[name].append(
                        {"effective": effective_date, "url": item["href"]}
                    )
            return packages
        else:
            return None

    def _extract_ct_name(self, url: str) -> str:
        match = re.search(r"([a-zA-Z]+)-\d{4}-\d{2}-\d{2}$", url)
        if match:
            return match.group(1)
        return None

    def _extract_effective_date(self, title: str) -> str:
        match = re.search(r"Effective (\d{4}-\d{2}-\d{2})$", title)
        if match:
            return match.group(1)
        return None

    def _url(self, relative_url: str) -> str:
        return f"{self.API_ROOT}{relative_url}"

    def _package_version(self, package: str) -> str | None:
        try:
            return self._packages[package][-1]["effective"]
        except Exception:
            return None



import os
from usdm3.bc.cdisc.library_api import LibraryAPI
from usdm3.bc.cdisc.library_cache.library_cache import LibraryCache
from usdm3.ct.cdisc.library import Library as CtLibrary


class Library:
    API_ROOT = "https://api.library.cdisc.org/api/cosmos/v2"
    BASE_PATH = "bc/cdisc"

    def __init__(self, root_path: str, ct_library: CtLibrary):
        self._ct_library = CtLibrary
        self.root_path = root_path
        self._api = LibraryAPI(ct_library)  # Interface to CDISC Library API
        self._cache = LibraryCache(
            os.path.join(self.root_path, self.BASE_PATH, "library_cache"),
            "library_cache.yaml",
        )  # Cache file handler
        self._bcs = {}
        self._bc_index = {}

    def load(self) -> None:
        if self._cache.exists():
            self._load_bcs()  # Load from cache file
        else:
            self._get_bcs()  # Fetch from API
            self._cache.save(self._bcs)  # Cache the results
        self._create_bc_index()

    def exists(self, name: str) -> bool:
        return True if name.upper() in self._bc_index else False

    def catalogue(self) -> list:
        return list(self._bcs.keys())

    def usdm(self, name: str) -> dict:
        return self._get_bc_data(name) if self.exists(name) else None

    def _load_bcs(self) -> None:
        self._bcs = self._cache.read()

    def _get_bcs(self) -> None:
        self._bcs = self._api.refresh()  # Ensure API connection is fresh

    def _create_bc_index(self) -> None:
        for name, item in self._bcs.items():
            self._bc_index[name] = name
            for synonym in item["synonyms"]:
                self._bc_index[synonym.upper()] = name

    def _get_bc_data(self, name: str) -> dict:
        return self._bcs[self._bc_index[name.upper()]]



from .rule_template import RuleTemplate


class RuleDDF00036(RuleTemplate):
    """
    DDF00036: If timing type is \"Fixed Reference\" then the corresponding attribute relativeToFrom must be filled with \"Start to Start\".

    Applies to: Timing
    Attributes: relativeToFrom
    """

    def __init__(self):
        super().__init__(
            "DDF00036",
            RuleTemplate.ERROR,
            'If timing type is "Fixed Reference" then the corresponding attribute relativeToFrom must be filled with "Start to Start".',
        )

    def validate(self, config: dict) -> bool:
        data = config["data"]
        items = data.instances_by_klass("Timing")
        for item in items:
            if item["type"]["decode"] == "Fixed Reference":
                if item["relativeToFrom"]["decode"] != "Start to Start":
                    self._add_failure(
                        "Invalid relativeToFrom",
                        "Timing",
                        "relativeToFrom",
                        data.path_by_id(item["id"]),
                    )
        return self._result()



from usdm3.api.wrapper import Wrapper
from usdm3.api.study import Study
from usdm3.api.study_title import StudyTitle
from usdm3.api.code import Code
from usdm3.api.study_protocol_document_version import StudyProtocolDocumentVersion
from usdm3.api.study_protocol_document import StudyProtocolDocument
from usdm3.api.study_version import StudyVersion
from usdm3.api.study_identifier import StudyIdentifier
from usdm3.api.organization import Organization
from usdm3.base.api_instance import APIInstance
from usdm3.base.id_manager import IdManager
from usdm3.ct.cdisc.library import Library
from usdm3.__info__ import __model_version__, __package_version__
from uuid import uuid4


class Minimum:
    @classmethod
    def minimum(
        cls, root_path: str, title: str, identifier: str, version: str
    ) -> "Wrapper":
        """
        Create a minimum study with the given title, identifier, and version.

        """
        api_classes = [
            Study,
            StudyTitle,
            StudyProtocolDocumentVersion,
            StudyProtocolDocument,
            StudyVersion,
            StudyIdentifier,
            Organization,
            Code,
            "Wrapper",
        ]

        id_manager = IdManager(api_classes)
        api_instance = APIInstance(id_manager)
        ct_library = Library(root_path)
        ct_library.load()
        cdisc_code_system = "cdisc.org"

        # Define the codes to be used in the study
        cl = ct_library.klass_and_attribute("StudyTitle", "type")
        title_type = api_instance.create(
            Code,
            {
                "code": "C207616",
                "codeSystem": cdisc_code_system,
                "codeSystemVersion": cl["source"]["effective_date"],
                "decode": "Official Study Title",
            },
        )
        cl = ct_library.klass_and_attribute(Organization, "organizationType")
        organization_type = api_instance.create(
            Code,
            {
                "code": "C70793",
                "codeSystem": cdisc_code_system,
                "codeSystemVersion": cl["source"]["effective_date"],
                "decode": "Clinical Study Sponsor",
            },
        )
        cl = ct_library.klass_and_attribute(
            StudyProtocolDocumentVersion, "protocolStatus"
        )
        doc_status = api_instance.create(
            Code,
            {
                "code": "C25425",
                "codeSystem": cdisc_code_system,
                "codeSystemVersion": cl["source"]["effective_date"],
                "decode": "Approved",
            },
        )

        study_title = api_instance.create(
            StudyTitle, {"text": title, "type": title_type}
        )

        # Define the protocol documents
        study_protocol_document_version = api_instance.create(
            StudyProtocolDocumentVersion,
            {"protocolVersion": version, "protocolStatus": doc_status},
        )
        study_protocol_document = api_instance.create(
            StudyProtocolDocument,
            {
                "name": "PROTOCOL",
                "label": "Study Protocol",
                "description": "The study protocol document",
                "versions": [study_protocol_document_version],
            },
        )

        # Define the organization and the study identifier
        organization = api_instance.create(
            Organization,
            {
                "name": "Sponsor",
                "organizationType": organization_type,
                "identifier": "To be provided",
                "identifierScheme": "To be provided",
                "legalAddress": None,
            },
        )
        study_identifier = api_instance.create(
            StudyIdentifier,
            {"studyIdentifier": identifier, "studyIdentifierScope": organization},
        )

        # Define the study version
        study_version = api_instance.create(
            StudyVersion,
            {
                "versionIdentifier": "1",
                "rationale": "To be provided",
                "titles": [study_title],
                "studyDesigns": [],
                "documentVersionId": study_protocol_document_version.id,
                "studyIdentifiers": [study_identifier],
            },
        )
        study = api_instance.create(
            Study,
            {
                "id": str(uuid4()),
                "name": "Study",
                "label": "",
                "description": "",
                "versions": [study_version],
                "documentedBy": study_protocol_document,
            },
        )

        # Return the wrapper for the study
        return api_instance.create(
            Wrapper,
            {
                "study": study,
                "usdmVersion": __model_version__,
                "systemName": "Python USDM3 Package",
                "systemVersion": __package_version__,
            },
        )



from typing import Literal, Union
from .api_base_model import ApiBaseModelNameLabelDesc
from .geographic_scope import SubjectEnrollment


class StudySite(ApiBaseModelNameLabelDesc):
    currentEnrollment: Union[SubjectEnrollment, None] = None
    instanceType: Literal["StudySite"]



import json
from simple_error_log.errors import Errors
from simple_error_log import ErrorLocation


class DataStoreErrorLocation(ErrorLocation):
    def __init__(self, path: str, klass: str, attribute: str):
        self.path = path
        self.klass = klass
        self.attribute = attribute

    def to_dict(self):
        return {"path": self.path, "klass": self.klass, "attribute": self.attribute}

    def __str__(self):
        return f"[{self.klass}', '{self.attribute}' @ '{self.path}']"


class DecompositionError(Exception):
    def __init__(self, error: DataStoreErrorLocation, message: str):
        self.error = error
        self.message = message

    def __str__(self):
        return f"error decomposing the '.json' file, {self.message}, at {self.error}"


class DataStore:
    def __init__(self, filename: str):
        self._klasses = {}
        self._ids = {}
        self._parent = {}
        self._path = {}
        self.filename = filename
        self.data = None
        self.errors = Errors()

    def decompose(self):
        self.data = self._load_data()
        self._check_study_id(self.data)
        self._decompose(self.data, None, "")

    def instance_by_id(self, id: str) -> dict:
        if id not in self._ids:
            return None
        return self._ids[id]

    def path_by_id(self, id: str) -> str:
        if id not in self._path:
            return None
        return self._path[id]

    def instances_by_klass(self, klass: str) -> list:
        if klass not in self._klasses:
            return []
        return list(self._klasses[klass].values())

    def parent_by_klass(self, id: str, klasses: str | list) -> dict:
        klasses = [klasses] if isinstance(klasses, str) else klasses
        found = False
        if id not in self._ids:
            return None
        instance = self._ids[id]
        while not found:
            if "instanceType" not in instance:
                return None
            elif instance["instanceType"] in klasses:
                found = True
            else:
                instance = self._parent[instance["id"]]
        return instance

    def _decompose(self, data, parent, path, instance_index=None) -> None:
        if isinstance(data, dict):
            path = self._add_klass_instance(data, parent, path, instance_index)
            for key, value in data.items():
                if isinstance(value, dict):
                    self._decompose(value, data, path)
                elif isinstance(value, list):
                    for index, item in enumerate(value):
                        self._decompose(item, data, path, index)

    def _add_klass_instance(self, data, parent, path, instance_index) -> None:
        id, klass = self._check_id_klass(parent, data, path)
        path = self._update_path(path, data, instance_index)
        if klass not in self._klasses:
            self._klasses[klass] = {}
        if id in self._ids:
            location = DataStoreErrorLocation(path, klass, "id")
            self.errors.add("Duplicate id '{id}' detected", location, "DUP_ID")
        self._klasses[klass][id] = data
        self._ids[id] = data
        self._path[id] = path
        self._parent[id] = parent
        return path

    def _load_data(self) -> dict:
        with open(self.filename, "r") as file:
            return json.load(file)

    def _update_path(self, path: str, data: dict, instance_index: int) -> str:
        path = path + "." + data["instanceType"] if "instanceType" in data else "$"
        path = path + f"[{instance_index}]" if instance_index is not None else path
        return path

    def _check_id_klass(self, parent: dict, data: dict, path: str) -> None:
        id, error = self._check_id(parent, data, path)
        if error:
            raise DecompositionError(error, "missing id attribute")
        klass, error = self._check_instance_type(parent, data, path)
        if error:
            raise DecompositionError(error, "missing instanceType attribute")
        return id, klass

    def _check_id(self, parent: dict, data: dict, path: str) -> None:
        if parent:
            if "id" in data:
                return data["id"], None
            else:
                klass = data["instanceType"] if "instanceType" in data else ""
                return None, DataStoreErrorLocation(path, klass, "id")
        else:
            return "$root", None

    def _check_instance_type(self, parent: dict, data: dict, path: str) -> None:
        if parent:
            if "instanceType" in data:
                return data["instanceType"], None
            else:
                id = data["id"] if "id" in data else ""
                return None, DataStoreErrorLocation(
                    path, f"missing instanceType for id '{id}'", "instanceType"
                )
        else:
            return "Wrapper", None

    def _check_study_id(self, data):
        # Do not want a null study id though it is permitted
        if "study" not in data:
            location = DataStoreErrorLocation("$", "study", "")
            raise DecompositionError(location, "missing study attribute")
        if "id" not in data["study"]:
            location = DataStoreErrorLocation("$.Study", "study", "id")
            raise DecompositionError(location, "missing id attribute")
        data["study"]["id"] = (
            "$root.study.id" if data["study"]["id"] is None else data["study"]["id"]
        )



from .rule_template import RuleTemplate


class RuleDDF00135(RuleTemplate):
    """
    DDF00135: An encounter's environmental setting must be specified according to the extensible Environmental Setting (C127262) SDTM codelist (e.g. an entry with a code or decode used from the codelist should be consistent with the full entry in the codelist).

    Applies to: Encounter
    Attributes: environmentalSetting
    """

    def __init__(self):
        super().__init__(
            "DDF00135",
            RuleTemplate.ERROR,
            "An encounter's environmental setting must be specified according to the extensible Environmental Setting (C127262) SDTM codelist (e.g. an entry with a code or decode used from the codelist should be consistent with the full entry in the codelist).",
        )



from typing import List, Literal
from .api_base_model import ApiBaseModelNameLabelDesc, ApiBaseModel


class ParameterMap(ApiBaseModel):
    tag: str
    reference: str
    instanceType: Literal["ParameterMap"]


class SyntaxTemplateDictionary(ApiBaseModelNameLabelDesc):
    parameterMaps: List[ParameterMap]
    instanceType: Literal["SyntaxTemplateDictionary"]
